import ProtocolEnum from '../enum/ProtocolEnum.js';
import ResourceLocatorIdentifierEnum from '../enum/ResourceLocatorIdentifierEnum.js';
/**
 *
 * The Resource Locator is a way for the nanotdf to represent references to external resources in as succinct a format
 * as possible.
 *
 * | Section       | Minimum Length (B) | Maximum Length (B) |
 * |---------------|--------------------|--------------------|
 * | Protocol Enum | 1                  | 1                  |
 * | Body Length   | 1                  | 1                  |
 * | Body          | 1                  | 255                |
 * | Identifier    | 0                  | n                  |
 *
 * @link https://github.com/virtru/nanotdf/blob/master/spec/index.md#3312-kas
 * @link https://github.com/virtru/nanotdf/blob/master/spec/index.md#341-resource-locator
 */
export default class ResourceLocator {
    readonly protocol: ProtocolEnum;
    readonly lengthOfBody: number;
    readonly body: string;
    readonly offset: number;
    readonly id?: string | undefined;
    readonly idType: ResourceLocatorIdentifierEnum;
    static readonly PROTOCOL_OFFSET = 0;
    static readonly PROTOCOL_LENGTH = 1;
    static readonly LENGTH_OFFSET = 1;
    static readonly LENGTH_LENGTH = 1;
    static readonly BODY_OFFSET = 2;
    static readonly IDENTIFIER_0_BYTE: number;
    static readonly IDENTIFIER_2_BYTE: number;
    static readonly IDENTIFIER_8_BYTE: number;
    static readonly IDENTIFIER_32_BYTE: number;
    constructor(protocol: ProtocolEnum, lengthOfBody: number, body: string, offset: number, id?: string | undefined, idType?: ResourceLocatorIdentifierEnum);
    /**
     * Construct a new URL or URL + identifier pair, for use with NanoTDF envelopes.
     * @param url The URL to encrypt; `http` and `https` schemes are supported
     * @param identifier An optional identifier.
     *    For KAS URLs, this is usually a public key identifier (kid). Limit 32 characters
     * @returns a value representing the URL and identifier, if present.
     *    This method throws an Error if the URL is invalid or of the wrong schema,
     *    or if the identifier is an unsupported value.
     */
    static fromURL(url: string, identifier?: string): ResourceLocator;
    static parse(buff: Uint8Array): ResourceLocator;
    /**
     * Length
     *
     * @returns { number } Length of resource locator
     */
    get length(): number;
    get url(): string | never;
    /**
     * Return the contents of the Resource Locator in buffer
     */
    toBuffer(): Uint8Array;
    /**
     * Get Identifier
     *
     * Returns the identifier of the ResourceLocator or an empty string if no identifier is present.
     * @returns { string } Identifier of the resource locator.
     */
    get identifier(): string;
}
//# sourceMappingURL=ResourceLocator.d.ts.map