import { KasPublicKeyAlgorithm, KasPublicKeyInfo, OriginAllowList } from '../../src/access.js';
import { type AuthProvider } from '../../src/auth/auth.js';
import { type Chunker } from '../../src/seekable.js';
import { AssertionConfig, AssertionVerificationKeys } from './assertions.js';
import { Binary } from './binary.js';
import { SymmetricCipher } from './ciphers/symmetric-cipher-base.js';
import { DecryptParams } from './client/builders.js';
import { DecoratedReadableStream } from './client/DecoratedReadableStream.js';
import { type CryptoService, type DecryptResult } from './crypto/declarations.js';
import { KeyAccessType, KeyInfo, Manifest, Policy, SplitKey, KeyAccess, KeyAccessObject, SplitType } from './models/index.js';
import { ZipReader } from './utils/index.js';
import { CentralDirectory } from './utils/zip-reader.js';
/**
 * Configuration for TDF3
 */
export type EncryptionOptions = {
    /**
     * Defaults to `split`, the currently only implmented key wrap algorithm.
     */
    type?: SplitType;
    cipher?: string;
};
type KeyMiddleware = DecryptParams['keyMiddleware'];
export type Metadata = unknown;
export type BuildKeyAccess = {
    type: KeyAccessType;
    alg?: KasPublicKeyAlgorithm;
    url?: string;
    kid?: string;
    publicKey: string;
    metadata?: Metadata;
    sid?: string;
};
type Mailbox<T> = Promise<T> & {
    set: (value: T) => void;
    reject: (error: Error) => void;
};
type Chunk = {
    hash: string;
    plainSegmentSize?: number;
    encryptedOffset: number;
    encryptedSegmentSize?: number;
    decryptedChunk: Mailbox<DecryptResult>;
};
export type IntegrityAlgorithm = 'GMAC' | 'HS256';
export type EncryptConfiguration = {
    allowList?: OriginAllowList;
    cryptoService: CryptoService;
    dpopKeys: CryptoKeyPair;
    encryptionInformation: SplitKey;
    segmentSizeDefault: number;
    integrityAlgorithm: IntegrityAlgorithm;
    segmentIntegrityAlgorithm: IntegrityAlgorithm;
    contentStream: ReadableStream<Uint8Array>;
    mimeType?: string;
    policy: Policy;
    authProvider?: AuthProvider;
    byteLimit: number;
    progressHandler?: (bytesProcessed: number) => void;
    keyForEncryption: KeyInfo;
    keyForManifest: KeyInfo;
    assertionConfigs?: AssertionConfig[];
    tdfSpecVersion?: string;
};
export type DecryptConfiguration = {
    allowedKases?: string[];
    allowList?: OriginAllowList;
    authProvider: AuthProvider;
    cryptoService: CryptoService;
    dpopKeys: CryptoKeyPair;
    chunker: Chunker;
    keyMiddleware: KeyMiddleware;
    progressHandler?: (bytesProcessed: number) => void;
    fileStreamServiceWorker?: string;
    assertionVerificationKeys?: AssertionVerificationKeys;
    noVerifyAssertions?: boolean;
    concurrencyLimit?: number;
    wrappingKeyAlgorithm?: KasPublicKeyAlgorithm;
};
export type UpsertConfiguration = {
    allowedKases?: string[];
    allowList?: OriginAllowList;
    authProvider: AuthProvider;
    privateKey: CryptoKey;
    unsavedManifest: Manifest;
    ignoreType?: boolean;
};
export type RewrapRequest = {
    signedRequestToken: string;
};
export type KasPublicKeyFormat = 'pkcs8' | 'jwks';
export type RewrapResponse = {
    entityWrappedKey: string;
    sessionPublicKey: string;
};
/**
 * If we have KAS url but not public key we can fetch it from KAS, fetching
 * the value from `${kas}/kas_public_key`.
 */
export declare function fetchKasPublicKey(kas: string, algorithm?: KasPublicKeyAlgorithm): Promise<KasPublicKeyInfo>;
export declare function extractPemFromKeyString(keyString: string, alg: KasPublicKeyAlgorithm): Promise<string>;
/**
 * Build a key access object and add it to the list. Can specify either
 * a (url, publicKey) pair (legacy, deprecated) or an attribute URL (future).
 * If all are missing then it attempts to use the default attribute. If that
 * is missing it throws an error.
 * @param  {Object} options
 * @param  {String} options.type - enum representing how the object key is treated
 * @param  {String} options.url - directly set the KAS URL
 * @param  {String} options.publicKey - directly set the (KAS) public key
 * @param  {String?} options.kid - Key identifier of KAS public key
 * @param  {String? Object?} options.metadata - Metadata. Appears to be dead code.
 * @return {KeyAccess}- the key access object loaded
 */
export declare function buildKeyAccess({ type, url, publicKey, kid, metadata, sid, alg, }: BuildKeyAccess): Promise<KeyAccess>;
export declare function validatePolicyObject(policy: Policy): void;
export declare function writeStream(cfg: EncryptConfiguration): Promise<DecoratedReadableStream>;
export type InspectedTDFOverview = {
    manifest: Manifest;
    zipReader: ZipReader;
    centralDirectory: CentralDirectory[];
};
export declare function loadTDFStream(chunker: Chunker): Promise<InspectedTDFOverview>;
export declare function splitLookupTableFactory(keyAccess: KeyAccessObject[], allowedKases: OriginAllowList): Record<string, Record<string, KeyAccessObject>>;
export declare function sliceAndDecrypt({ buffer, reconstructedKeyBinary, slice, cipher, segmentIntegrityAlgorithm, specVersion, }: {
    buffer: Uint8Array;
    reconstructedKeyBinary: Binary;
    slice: Chunk[];
    cipher: SymmetricCipher;
    cryptoService: CryptoService;
    segmentIntegrityAlgorithm: IntegrityAlgorithm;
    specVersion: string;
}): Promise<void>;
export declare function readStream(cfg: DecryptConfiguration): Promise<DecoratedReadableStream>;
export declare function decryptStreamFrom(cfg: DecryptConfiguration, { manifest, zipReader, centralDirectory }: InspectedTDFOverview): Promise<DecoratedReadableStream>;
export {};
//# sourceMappingURL=tdf.d.ts.map